#include "tg_io.h"

/* ------------------------------------------------------------------------ */
/*
 * Simple interfaces to the underlying g-library. These are very basic, but
 * sufficient. Will need to rework this with proper lock utilisation schemes
 * for Gap4 proper.
 */

/*
 * Ripped out of Gap4's gap-create.c
 *
 * create a new file header
 * returns:    0 - ok
 *          else - error
 */
static int bitsize = G_64BIT;
int create_new_files(char *fn, size_t s, GCardinal max_rec)
{
    char auxfn[1024];
    int fd;
    AuxHeader auxheader;
    GCardinal i;
    int (*(*low_level_vector))(int fd, void *x, int num);
    int endian = 1;

    /*
     * Determine the default vectors for creating a database.
     * This is temporary as from here on the g-library will auto-sense when
     * it opens the database files.
     */
    if ( *(char *)&endian ) {
	low_level_vector = (bitsize == G_64BIT)
	    ? low_level_vectors_swapped64
	    : low_level_vectors_swapped32;
    } else {
	low_level_vector = (bitsize == G_64BIT)
	    ? low_level_vectors64
	    : low_level_vectors32;
    }

    /* check file name isn't too long */
    if ( strlen(fn) + strlen(G_AUX_SUFFIX) >= sizeof(auxfn) ) return gerr_set(GERR_NAME_TOO_LONG);
	
    strcpy(auxfn,fn);
    strcat(auxfn,G_AUX_SUFFIX);

    /* create files */
    /* LOW LEVEL IO HERE */
    if ( (fd = creat(fn,G_DEF_PERMS)) == -1 ) return gerr_set(GERR_CANT_CREATE);
    /* LOW LEVEL IO HERE */
    close(fd);

    /* LOW LEVEL IO HERE */
    if ( (fd = creat(auxfn,G_DEF_PERMS)) == -1 ) return gerr_set(GERR_CANT_CREATE);

    /* initialise header */
    auxheader.file_size = 0;
    auxheader.block_size = (int4) s;
    auxheader.num_records = 0;
    auxheader.max_records = max_rec;
    auxheader.last_time = G_YEAR_DOT;
    auxheader.flags = (GFlags) 0;
    auxheader.spare1 = (GFlags) 0;
    auxheader.free_time = G_YEAR_DOT;
    for (i=G_Number(auxheader.spare)-1;i>=0;i--) auxheader.spare[i]=0;
    auxheader.format = bitsize;

    /* write(fd,&auxheader,sizeof(auxheader)); */
    (void) (low_level_vector[GOP_WRITE_AUX_HEADER])(fd,&auxheader,1);

    /* LOW LEVEL IO HERE */
    close(fd);
    return 0;
}

/* Hacky allocation scheme - always increment the record number */
int allocate(IO *io) {
    static int record = 0;
    return record++;
}

GView lock(IO *io, int rec) {
    return g_lock_N_(io->db, io->client, 0, rec, G_LOCK_EX);
}

int unlock(IO *io, GView v) {
    return g_unlock_(io->db, io->client, v);
}

IO *g_open(char *fn, int ro) {
    IO *io = (IO *)calloc(1, sizeof(*io));

    io->db = g_open_database_(&fn, 1, ro);
    io->client = g_connect_client_(io->db, 0, G_LOCK_EX, &io->mode);
    return io;
}

void g_close(IO *io) {
    g_disconnect_client_(io->db, io->client);
    g_shutdown_database_(io->db);
}

void g_free(IO *io) {
    g_close(io);
    free(io);
}

int g_write(IO *io, GView v, void *buf, size_t len) {
    return g_write_(io->db, io->client, v, buf, len);
}

/*
 * Writes buf as a series of little-endian int32_t elements, byte
 * swapping if appropriate.
 */
int g_write_le4(IO *io, GView v, void *buf, size_t len) {
    static GCardinal le_a[256];
    int32_t *le = le_a;
    int ret;
    size_t i, j;

    if (len % sizeof(int32_t) != 0)
	return -1;

    if (len > 256 * sizeof(int32_t))
	le = (int32_t *)malloc(len * sizeof(int32_t));

    for (i = j = 0; i < len; i+=sizeof(int32_t), j++) {
	le[j] = le_int4(((int32_t *)buf)[j]);
    }

    ret = g_write_(io->db, io->client, v, le, len);

    if (le != le_a)
	free(le);
    return ret;
}

int g_read(IO *io, GView v, void *buf, size_t len) {
    return g_read_(io->db, io->client, v, buf, len);
}

void *g_read_alloc(IO *io, GView v, size_t *len) {
    GViewInfo vi;
    void *buf;

    g_view_info_(io->db, io->client, v, &vi);
    if (len)
	*len = vi.used;

    if (NULL == (buf = malloc(vi.used)))
	return NULL;
    
    if (g_read_(io->db, io->client, v, buf, vi.used) == 0)
	return buf;
    else {
	free(buf);
	return NULL;
    }
}

/*
 * Reads buf as a series of little-endian int32_t elements, byte
 * swapping if appropriate.
 */
int g_read_le4(IO *io, GView v, void *buf, size_t len) {
    int ret;
    size_t i, j;
    int32_t *le = (int32_t *)buf;

    if (len % sizeof(int32_t) != 0)
	return -1;

    ret = g_read_(io->db, io->client, v, buf, len);

    for (i = j = 0; i < len; i+=sizeof(int32_t), j++) {
	le[j] = le_int4(((int32_t *)buf)[j]);
    }

    return ret;
}

int g_flush(IO *io, GView v) {
    return g_flush_(io->db, io->client, v);
}

